@@ -8,19 +8,28 @@ use Scalar::Util 'reftype';
 use POSIX ':sys_wait_h';
 require overload;
 
+# TODO: more expressive "security" policy, which must now cover whether
+#       advisory locks may be stolen.
+
+our $archy_avail = eval "use Patro::Archy;1";
+#$Patro::Archy::DEBUG = 1;
 our $threads_avail;
 *sxdiag = sub {};
 if ($ENV{PATRO_SERVER_DEBUG}) {
     *sxdiag = *::xdiag;
     our $DEBUG = 1;
 }
-our $VERSION = '0.14';
+our $VERSION = '0.16';
 our @SERVERS :shared;
-our %OPTS = ( # XXX - needs documentation
-    keep_alive => 30,
-    idle_timeout => 30,
-    fincheck_freq => 5,
-);
+our %OPTS;
+
+$OPTS{keep_alive} //= 30;
+$OPTS{idle_timeout} //= 30;
+$OPTS{fincheck_freq} //= 5;
+
+$OPTS{secure_filehandles} //= 1;
+$OPTS{steal_lock_ok} //= 1;
+
 
 sub new {
     my $pkg = shift;
@@ -112,7 +121,7 @@ sub new {
     }, __PACKAGE__;
     $self->{config} = $self->config;
     $self->start_server;
-    eval { push @SERVERS, $self };
+    push @SERVERS, threadsx::shared::_shared_clone($self);
     warn $@ if $@;
     if (@SERVERS == 1) {
 	eval q~END {
@@ -296,12 +305,20 @@ sub accept_clients {
 
 sub start_subserver {
     my ($self,$client) = @_;
+    # monitor id is not shared across forks or threads,
+    # so each server will have a unique value
+
+    our $monitor_id;
+    $monitor_id //= 0;
+    $monitor_id = sprintf "%010X", hex($monitor_id) + 10 + 1014 * rand;
+
     if ($self->{meta}{style} eq 'forked') {
 	my $pid = CORE::fork();
 	if (!defined($pid)) {
 	    croak __PACKAGE__,": fork after accept $!";
 	}
 	if ($pid != 0) {
+	    $monitor_id = sprintf "%010X", hex($monitor_id) + 5 + 256 * rand;
 	    if ($self->{meta}{pid_file}) {
 		open my $fh, '>>', $self->{meta}{pid_file};
 		flock $fh, 2;
@@ -355,10 +372,17 @@ sub watch_for_finishers {
 	if (@joinable) {
 	    foreach my $subthread  (@joinable) {
 		my ($i) = grep {
-		    $self->{meta}{subthreads}{$_} == $subthread 
+		    my $is = eval {$self->{meta}{subthreads}{$_} == $subthread};
+		    if ($@ && 0) {
+			::xdiag("Server: exception checking for finishers: ",
+				$@,"\n",$self->{meta},$self->{meta}{subthreads});
+		    }
+		    $is;
 		} 0 .. $n-1;
 		if (!defined($i)) {
-		    warn "subthread $subthread not found on this server!";
+		    # this is not unusual if the Patro user is using threads
+#		    warn "subthread $subthread ",0+$subthread,
+#			" not found on this server!";
 		    next;
 		}
 		$self->{meta}{subthreads}[$i]->join;
@@ -468,6 +492,8 @@ sub process_request {
 	@r = $self->process_request_CODE($id,undef,$ctx,$has_args,$args);
     } elsif ($topic eq 'HANDLE') {
 	@r = $self->process_request_HANDLE($id,$cmd,$ctx,$has_args,$args);
+    } elsif ($topic eq 'SYNC') {
+	@r = $self->process_request_SYNC($id,$cmd,$ctx,$has_args,$args);
     } elsif ($topic eq 'OVERLOAD') {
 	my $obj = $self->{obj}{$id};
 	@r = $self->process_request_OVERLOAD($obj,$cmd,$args,$ctx);
@@ -486,7 +512,6 @@ sub process_request {
     $sides->{errno_extended} = $^E if $^E;
     $sides->{child_error} = $? if $?;
     $sides->{error} = $@ if $@;
-    $sides->{"x-requestId"} = ++$Patro::Server::requestId;
 
     # how to update elements of @_ that have changes?
     # three implementations below. Pick one.
@@ -552,6 +577,73 @@ sub process_request_META {
     }
 }
 
+sub process_request_SYNC {
+    my ($self,$id,$cmd,$ctx,$has_args,$args) = @_;
+    if (!$archy_avail) {
+	$! = 999;
+	$SIDES->{warn} = "Patro synchronization feature not available";
+	return;
+    }
+    my $obj = $self->{obj}{$id};
+    if (!defined($obj)) {
+	our $SIDES;
+	$SIDES->{_fatal} = "Server: object id $id is not associated "
+	    . "with a known reference\n"
+	    . "Known ids: " . join(" ",sort keys %{$self->{obj}});
+	return;
+    }
+    our $monitor_id;
+    if ($cmd eq 'lock') {
+	our $SIDES;
+	if ($has_args) {
+	    if (!$OPTS{steal_lock_ok}) {
+		if ($args->[1]) {
+		    $SIDES->{warn} = "Patro::lock: steal=true not allowed "
+				   . "by server";
+		    $args->[1] = 0;
+		}
+	    }
+	    return Patro::Archy::plock($obj, $monitor_id, @$args);
+	} else {
+	    return Patro::Archy::plock($obj, $monitor_id);
+	}
+    } elsif ($cmd eq 'unlock') {
+	if ($has_args) {
+	    return Patro::Archy::punlock($obj, $monitor_id, $args->[0]);
+	} else {
+	    return Patro::Archy::punlock($obj, $monitor_id);
+	}
+    } elsif ($cmd eq 'wait') {
+	if ($has_args) {
+	    return Patro::Archy::pwait($obj,$monitor_id,$args->[0]);
+	} else {
+	    return Patro::Archy::pwait($obj,$monitor_id);
+	}
+    } elsif ($cmd eq 'notify') {
+	if ($has_args) {
+	    return Patro::Archy::pnotify($obj,$monitor_id,$args->[0]);
+	} else {
+	    return Patro::Archy::pnotify($obj,$monitor_id);
+	}
+    } elsif ($cmd eq 'state') {
+	my $state = Patro::Archy::pstate($obj, $monitor_id);
+	if ($state == Patro::Archy::STATE_NULL()) {
+	    return "0,NULL";
+	} elsif ($state == Patro::Archy::STATE_WAIT()) {
+	    return "1,WAIT";
+	} elsif ($state == Patro::Archy::STATE_NOTIFY()) {
+	    return "2,NOTIFY";
+	} elsif ($state == Patro::Archy::STATE_STOLEN()) {
+	    return "3,STOLEN";
+	} else {
+	    return "$state,LOCK";
+	}
+    } else {
+	$@ = "Unrecognized SYNC command '$cmd'";
+	return;
+    }
+}
+
 sub process_request_HASH {
     my ($self,$id,$cmd,$ctx,$has_args,$args) = @_;
     my $obj = $self->{obj}{$id};
@@ -566,7 +658,8 @@ sub process_request_HASH {
 	$obj->{$key} = threads::shared::shared_clone($val);
 	return $old_val;
     } elsif ($cmd eq 'FETCH') {
-	return $obj->{$args->[0]};
+	my $key = $args->[0];
+	return $obj->{$key};
     } elsif ($cmd eq 'DELETE') {
 	return delete $obj->{$args->[0]};
     } elsif ($cmd eq 'EXISTS') {
@@ -780,14 +873,14 @@ sub process_request_HANDLE {
 	}
 	return $z;
     } elsif ($command eq 'CLOSE') {
-	if ($Patro::SECURE) {
+	if ($OPTS{secure_filehandles}) {
 	    $@ = "Patro: insecure CLOSE operation on proxy filehandle";
 	    return;
 	}
 	my $z = close $fh;
 	return $z;    
     } elsif ($command eq 'OPEN') {
-	if ($Patro::SECURE) {
+	if ($OPTS{secure_filehandles}) {
 	    $@ = "Patro: insecure OPEN operation on proxy filehandle";
 	    return;
 	}
@@ -803,7 +896,7 @@ sub process_request_HANDLE {
 		$z = open $fh, $mode, $expr, @$args;
 	    }
 	}
-
+	
 	# it is hard to set autoflush from the proxy.
 	# Since it is usually what you want, let's do it here.
 	if ($z) {
@@ -837,19 +930,24 @@ sub process_request_HANDLE {
 	}
     } elsif ($command eq '-X') {
 	my $key = $args->[0];
+	# insecure unless $key is vetted
+	if (length($key) != 1 || $key !~ /[A-Za-z]/) {
+	    $@ = "Invalid filetest operator '-$key'";
+	    return;
+	}
 	return eval "-$key \$fh";
     } elsif ($command eq 'SYSOPEN') {
-	if ($Patro::SECURE) {
+	if ($OPTS{secure_filehandles}) {
 	    $@ = "Patro: insecure SYSOPEN operation on proxy filehandle";
 	    return;
 	}
         my $z = @$args <= 2 ? sysopen $fh, $args->[0], $args->[1]
-                      : sysopen $fh, $args->[0], $args->[1], $args->[2];
+                            : sysopen $fh, $args->[0], $args->[1], $args->[2];
         return $z;
 
     # commands that operate on DIRHANDLEs
     } elsif ($command eq 'OPENDIR') {
-        if ($Patro::SECURE) {
+        if ($OPTS{secure_filehandles}) {
             $@ = "Patro: insecure OPENDIR operation on proxy dirhandle";
             return;
         }
@@ -893,14 +991,22 @@ sub process_request_CODE {
 
 sub process_request_OVERLOAD {
     my ($self,$x,$op,$args,$context) = @_;
-    if ($op eq '@{}') { 
-	return \@$x;
+    if ($op eq '@{}') {
+	my $z = eval { \@$x };
+	$@ &&= "Not an ARRAY reference";
+	return $z;
     } elsif ($op eq '%{}') {
-	return \%$x;
+	my $z = eval { \%$x };
+	$@ &&= "Not a HASH reference";
+	return $z;
     } elsif ($op eq '&{}') {
-	return \&$x;
+	my $z = eval { \&$x };
+	$@ &&= "Not a CODE reference";
+	return $z;
     } elsif ($op eq '${}') {
-	return \$$x;
+	my $z = eval { \$$x };
+	$@ &&= "Not a SCALAR reference";
+	return $z;
     } # elsif ($op eq '*{}') { return \*$x; }
     my ($y,$swap) = @$args;
     if ($swap) {
@@ -912,10 +1018,16 @@ sub process_request_OVERLOAD {
         $op =~ s/\.//;
     }
     if ($op eq '-X') {
+	# !!! insecure unless $y is vetted
+	if (length($y) != 1 || $y !~ /[A-Za-z]/) {
+	    $@ = "Invalid filetest operator '$y'";
+	    return;
+	}
         $z = eval "-$y \$x";
     } elsif ($op eq 'neg') {
         $z = eval { -$x };
     } elsif ($op eq '!' || $op eq '~' || $op eq '++' || $op eq '--') {
+	# string eval ok -- $op is vetted
         $z = eval "$op\$x";
     } elsif ($op eq 'qr') {
         $z = eval { qr/$x/ };
@@ -923,6 +1035,7 @@ sub process_request_OVERLOAD {
         $z = eval { atan2($x,$y) };
     } elsif ($op eq 'cos' || $op eq 'sin' || $op eq 'exp' || $op eq 'abs' ||
              $op eq 'int' || $op eq 'sqrt' || $op eq 'log') {
+	# string eval ok -- $op is vetted
         $z = eval "$op(\$x)";
     } elsif ($op eq 'bool') {
         $z = eval { $x ? 1 : 0 };  # this isn't right
@@ -934,6 +1047,15 @@ sub process_request_OVERLOAD {
         # always scalar context readline
         $z = eval { readline($x) };
     } else {  # binary operator
+	# !!! insecure unless $op is vetted
+	if (!@oplist) {
+	    @oplist = split ' ',join(' ',values %overload::ops);
+	}
+	my %ops = map {; $_ => 1 } @oplist;
+	if (!defined $ops{$op}) {
+	    $@ = "Invalid operator '$op'";
+	    return;
+	}
         $z = eval "\$x $op \$y";
     }
     if ($@) {
@@ -998,18 +1120,18 @@ sub serialize_response {
     my ($self, $resp) = @_;
     if ($resp->{context}) {
 	if ($resp->{context} == 1) {
-	    $resp->{response} = patrol($self,$resp,$resp->{response});
+	    $resp->{response} = patrofy($self,$resp,$resp->{response});
 	} elsif ($resp->{context} == 2) {
 	    $resp->{response} = [
-		map patrol($self,$resp,$_), @{$resp->{response}} ];
+		map patrofy($self,$resp,$_), @{$resp->{response}} ];
 	}
     }
 
     if ($resp->{out}) {
-	$resp->{out} = [ map patrol($self,$resp,$_), @{$resp->{out}} ];
+	$resp->{out} = [ map patrofy($self,$resp,$_), @{$resp->{out}} ];
     }
 
-    sxdiag("Server: final response before serialization: ",$resp);
+    sxdiag("Patro::Server: final response before serialization: ",$resp);
     $resp = Patro::LeumJelly::serialize($resp);
     return $resp;
 }
@@ -1017,18 +1139,18 @@ sub serialize_response {
 # we should not send any serialized references back to the client.
 # replace any references in the response with an
 # object id.
-sub patrol {
+sub patrofy {
     my ($self,$resp,$obj) = @_;
-    sxdiag("patrol: called on: ",defined($obj) ? "$obj" : "<undef>");
+    sxdiag("patrofy: called on: ",defined($obj) ? "$obj" : "<undef>");
     return $obj unless ref($obj);
 
     if ($threads_avail) {
 	if (CORE::ref($obj) eq 'CODE') {
 	    $obj = threadsx::shared::code->new($obj);
-	    sxdiag("patrol: coderef converted");
+	    sxdiag("patrofy: coderef converted");
 	} elsif (CORE::ref($obj) eq 'GLOB') {
 	    $obj = threadsx::shared::glob->new($obj);
-	    sxdiag("patrol: glob converted");
+	    sxdiag("patrofy: glob converted");
 	}
     }
 
@@ -1050,7 +1172,7 @@ sub patrol {
 	} else {
 	    $reftype = Scalar::Util::reftype($obj);
 	}
-	sxdiag("patrol: ref types for $id are $ref,$reftype");
+	sxdiag("patrofy: ref types for $id are $ref,$reftype");
 	$resp->{meta}{$id} = {
 	    id => $id, ref => $ref, reftype => $reftype
 	};
@@ -1089,7 +1211,7 @@ Patro::Server - remote object server for Patro
 
 =head1 VERSION
 
-0.14
+0.16
 
 =head1 DESCRIPTION
 
@@ -1099,6 +1221,91 @@ The server handles requests for any references that are being served,
 manipulates references on the server, and returns the results of
 operations to the proxy objects on the clients.
 
+=head1 SERVER OPTIONS
+
+Server options are configured through the global
+C<%Patro::Server::OPTS> hash. Changes to this hash generally
+only affect servers in threads and processes that are started
+after the changes are made; any desired changes should be
+made before the first call to L<Patro/"patronize">.
+
+The server recognizes the following options in the
+C<%Patro::Server::OPTS> hash:
+
+=head2 Server operation options
+
+In its original incarnation, I imagined that the servers and
+clients would be tightly coupled. A server would start, and
+then the same person that started the server would start one
+or more clients on other machines in the same cluster. The
+clients would perform some tasks, and then exit. When the last
+client disconnected, and after a little bit of waiting to see
+if any new clients would connect, the server would shut down.
+Other server use cases are possible, but this is the use case
+that informed these server settings.
+
+=over 4
+
+=item keep_alive => seconds
+
+After the server starts, this is the number of seconds it is
+guaranteed to stay up, even if no clients connect to it.
+The default is 30 seconds; you could set this to a very
+large value in order to have a persistent server.
+
+=item idle_timeout => seconds
+
+After all clients have disconnected, this is the number of
+seconds to wait for another client to connect. If another
+client does not connect before this time expires, then the
+server will shut down (also subject to satisfying the
+C<keep_alive> timeout, described above). The default is
+30 seconds.
+
+=item fincheck_freq => seconds
+
+The number of seconds in between checks of which clients
+are still active. When no active clients are detected,
+the C<idle_timeout> count down will begin.
+
+This check is implemented with a C<SIGALRM>. That may be
+inconvenient to the many other programs and modules that
+would use C<SIGALRM> though, so this may change in the future.
+
+=back
+
+=head2 Server security options
+
+These options indicate whether certain operations that may
+be insecure or impolite are allowed on the server.
+
+=over 4
+
+=item secure_filehandles => bool
+
+If this option is true, then proxies to filehandles will
+receive an error if they execute the C<open> or C<close>
+functions on the filehandles. If the C<open> function is
+allowed, the client will have read/write access (with
+the userid of the server process) to any file on the
+server filesystem. Worse, the client could use a pipe
+open (C<open $proxy_fh, "| arbitrary command">) and
+execute arbitrary code on the server. Calling C<close>
+on a proxy filehandle is not so dangerous but it would still
+not be allowed if this option is set.
+
+=item steal_lock_ok => bool
+
+C<Patro> provides advisory locking on remote references.
+One option to the L<Patro::lock|Patro/"lock"> function is
+to allow a lock to be stolen from another monitor. Whether
+stealing locks should be allowed depends a lot on your
+application. Some reasons you might want to allow it is if
+you have slow and unreliable processes or if your application
+is prone to deadlock.
+
+=back
+
 =head1 LICENSE AND COPYRIGHT
 
 MIT License
